/*
 * VeriSilicon sample for decoding with VPE suppprted avcodec API.
 * Copyright (C) 2020 VeriSilicon Holdings Co., Ltd.
 *
 * This file is part of FFmpeg.
 *
 * FFmpeg is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * FFmpeg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FFmpeg; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include <sys/time.h>
#include <unistd.h>
#include <semaphore.h>
#include <pthread.h>

#include <libavcodec/avcodec.h>
#include <libavcodec/internal.h>
#include <libavformat/avformat.h>
#include <libavfilter/buffersink.h>
#include <libavfilter/buffersrc.h>
#include <libavutil/opt.h>
#include <libavutil/log.h>
#include <libavutil/pixdesc.h>
#include <libavutil/hwcontext.h>
#include <libavutil/avutil.h>

static const int avloglevel = AV_LOG_DEBUG;
static const char *device   = "/dev/ama_transcoder0";
static char *dec_name;
static char *output_file;

static const char *device_option_string = "priority=vod,vpeloglevel=6";
const char *filter_descr = "hwdownload,format=nv12";

static AVBufferRef *hw_device_ctx = NULL;

static AVFormatContext *fmt_ctx;
static AVCodecContext *dec_ctx = NULL;
AVFilterContext *buffersink_ctx;
AVFilterContext *buffersrc_ctx;
AVFilterGraph *filter_graph;
static int video_stream_index = -1;
static int frame_out_cnt;
struct timeval start_time;
static int filter_init = 0;
static int output_cnt = 1;

static FILE *fp                   = NULL;

static int vpe_get_format(AVCodecContext *avctx, const enum AVPixelFormat *pix_fmts)
{
    if (hw_device_ctx)
        return AV_PIX_FMT_VPE;
    else
        return AV_PIX_FMT_NONE;
}

static int open_input_file(const char *filename)
{
    int i;
    int ret;
    AVCodec *dec;
    AVDictionary *opts = NULL;
    AVStream *stream;

    dec = avcodec_find_decoder_by_name(dec_name);
    if (!dec) {
        av_log(NULL, AV_LOG_FATAL, "Necessary decoder not found %s\n", dec_name);
        return -1;
    }

    if ((ret = avformat_open_input(&fmt_ctx, filename, NULL, &opts)) < 0) {
        av_log(NULL, AV_LOG_ERROR, "Cannot open input file\n");
        return ret;
    }

    if ((ret = avformat_find_stream_info(fmt_ctx, NULL)) < 0) {
        av_log(NULL, AV_LOG_ERROR, "Cannot find stream information\n");
        return ret;
    }
    video_stream_index = 0;

    for (i = 0; i < fmt_ctx->nb_streams; i++) {
        if (fmt_ctx->streams[i]->codecpar->codec_type == AVMEDIA_TYPE_VIDEO) {
            video_stream_index = i;
            av_log(NULL, AV_LOG_DEBUG, "get video stream index %d\n", i);
            break;
        }
    }
    /* create decoding context */
    stream = fmt_ctx->streams[video_stream_index];
    dec_ctx = avcodec_alloc_context3(dec);
    if (!dec_ctx)
        return AVERROR(ENOMEM);
    avcodec_parameters_to_context(dec_ctx, stream->codecpar);
    dec_ctx->hw_device_ctx = av_buffer_ref(hw_device_ctx);
    dec_ctx->get_format = vpe_get_format;

    if (output_cnt != 1) {
        /* init the video decoder */
        if ((ret = avcodec_open2(dec_ctx, dec, &opts)) < 0) {
            av_log(NULL, AV_LOG_ERROR, "Cannot open video decoder\n");
            return ret;
        }
    } else {
        /* init the video decoder */
        if ((ret = avcodec_open2(dec_ctx, dec, NULL)) < 0) {
            av_log(NULL, AV_LOG_ERROR, "Cannot open video decoder\n");
            return ret;
        }
    }


    return 0;
}

static int init_filters(const char *filters_spec)
{
    char args[512];
    int ret                         = 0;
    const AVFilter *buffersrc       = NULL;
    const AVFilter *buffersink      = NULL;
    AVFilterInOut *outputs          = avfilter_inout_alloc();
    AVFilterInOut *inputs           = avfilter_inout_alloc();
    AVFilterGraph *filter_graph     = avfilter_graph_alloc();

    AVRational time_base = fmt_ctx->streams[video_stream_index]->time_base;
    int i;

    if (!outputs || !inputs || !filter_graph) {
        ret = AVERROR(ENOMEM);
        goto end;
    }

    av_log(NULL, AV_LOG_DEBUG, "filters_spec '%s' for stream\n", filters_spec);

    buffersrc  = avfilter_get_by_name("buffer");
    buffersink = avfilter_get_by_name("buffersink");
    if (!buffersrc || !buffersink) {
        av_log(NULL, AV_LOG_ERROR,
               "filtering source or sink element not found\n");
        ret = AVERROR_UNKNOWN;
        goto end;
    }

    snprintf(args, sizeof(args),
             "video_size=%dx%d:pix_fmt=%d:time_base=%d/%d:pixel_aspect=%d/%d",
             dec_ctx->width, dec_ctx->height, dec_ctx->pix_fmt,
             time_base.num, time_base.den,
             dec_ctx->sample_aspect_ratio.num,
             dec_ctx->sample_aspect_ratio.den);
    printf("args %s\n", args);

    ret = avfilter_graph_create_filter(&buffersrc_ctx, buffersrc, "in",
                                       args, NULL, filter_graph);
    if (ret < 0) {
        av_log(NULL, AV_LOG_ERROR, "Cannot create buffer source\n");
        goto end;
    }
    ret = avfilter_graph_create_filter(&buffersink_ctx, buffersink, "out",
                                       NULL, NULL, filter_graph);
    if (ret < 0) {
        av_log(NULL, AV_LOG_ERROR, "Cannot create buffer sink\n");
        goto end;
    }

    /* Endpoints for the filter graph. */
    outputs->name       = av_strdup("in");
    outputs->filter_ctx = buffersrc_ctx;
    outputs->pad_idx    = 0;
    outputs->next       = NULL;

    inputs->name       = av_strdup("out");
    inputs->filter_ctx = buffersink_ctx;
    inputs->pad_idx    = 0;
    inputs->next       = NULL;

    if (!outputs->name || !inputs->name) {
        ret = AVERROR(ENOMEM);
        goto end;
    }

    av_log(NULL, AV_LOG_DEBUG, "to avfilter_graph_parse_ptr\n");

    if ((ret = avfilter_graph_parse_ptr(filter_graph, filters_spec, &inputs,
                                        &outputs, NULL)) < 0)
        goto end;

    av_log(NULL, AV_LOG_DEBUG, "to set hwdevice to filter\n");

    if (dec_ctx->codec_type == AVMEDIA_TYPE_VIDEO && hw_device_ctx) {
        for (i = 0; i < filter_graph->nb_filters; i++) {
            filter_graph->filters[i]->hw_device_ctx =
                av_buffer_ref(hw_device_ctx);
            if (!filter_graph->filters[i]->hw_device_ctx) {
                ret = AVERROR(ENOMEM);
                goto end;
            }
            printf("%d nb_inputs %d\n", i, filter_graph->filters[i]->nb_inputs);
            printf("%d nb_outputs %d\n", i, filter_graph->filters[i]->nb_outputs);
            printf("%d, name %s\n", i, filter_graph->filters[i]->name);
            if (strstr(filter_graph->filters[i]->name, "out"))
                printf("input %p\n", filter_graph->filters[i]->inputs[0]);
            if (strstr(filter_graph->filters[i]->name, "hwdownload") ||
                strstr(filter_graph->filters[i]->name, "spliter")) {
                printf("%d, hw_frames %p\n", i, dec_ctx->hw_frames_ctx);
                printf("input %p\n", filter_graph->filters[i]->inputs[0]);
                filter_graph->filters[i]->inputs[0]->hw_frames_ctx =
                    av_buffer_ref(dec_ctx->hw_frames_ctx);
                printf("%d line\n", __LINE__);
            }
        }
    }
    av_log(NULL, AV_LOG_DEBUG, "to avfilter_graph_config\n");
    if ((ret = avfilter_graph_config(filter_graph, NULL)) < 0)
        goto end;
    av_log(NULL, AV_LOG_DEBUG, "to avfilter_graph_config done\n");

end:
    avfilter_inout_free(&inputs);
    avfilter_inout_free(&outputs);

    return ret;
}

static int init_vpe_device(void)
{
    int ret;
    AVDictionary *opts = NULL;

    av_dict_parse_string(&opts, device_option_string, "=", ",", 0);
    ret = av_hwdevice_ctx_create(&hw_device_ctx, AV_HWDEVICE_TYPE_AMA, device,
                                 opts, 0);
    if (ret < 0) {
        av_log(NULL, AV_LOG_ERROR,
               "Failed to create a VPE device. Error code: %s\n",
               av_err2str(ret));
        return -1;
    }
    return 0;
}

static int store_frame(AVFrame *frame)
{
    int ret = 0;
    int i;

    for (i = 0; i < frame->height; i++) {
        if ((ret = fwrite((frame->data[0]+i*frame->linesize[0]),
                           1, frame->linesize[0], fp)) < 0) {
            printf("Failed to dump raw data.\n");
            return -1;
        }
    }

    for (i = 0; i < frame->height/2; i++) {
        if ((ret = fwrite((frame->data[1]+i*frame->linesize[1]),
                           1, frame->linesize[1], fp)) < 0) {
            printf("Failed to dump raw data.\n");
            return -1;
        }
    }

    fflush(fp);
    return 0;
}

static int dec_video(AVPacket *pkt, AVFrame *frame, AVFrame *filt_frame)
{
    int ret, sink_ret;

    if (pkt->stream_index == video_stream_index) {
        ret = avcodec_send_packet(dec_ctx, pkt);
        if (ret < 0) {
            av_log(NULL, AV_LOG_ERROR, "Error while sending a packet to the decoder, ret %d\n", ret);
            return -1;
        }

        while (ret >= 0) {
            ret = avcodec_receive_frame(dec_ctx, frame);
            if (ret == AVERROR(EAGAIN) || ret == AVERROR_EOF) {
                break;
            } else if (ret < 0) {
                av_log(NULL, AV_LOG_ERROR, "Error while receiving a frame from the decoder\n");
                return -1;
            }

            if (!filter_init) {
                if ((ret = init_filters(filter_descr)) < 0)
                    return -1;
                filter_init = 1;
            }

            frame->pts = frame->best_effort_timestamp;

            /* push the decoded frame into the filtergraph */
            if (av_buffersrc_add_frame_flags(buffersrc_ctx, frame, AV_BUFFERSRC_FLAG_KEEP_REF) < 0) {
                av_log(NULL, AV_LOG_ERROR, "Error while feeding the filtergraph\n");
                break;
            }

            /* pull filtered frames from the filtergraph */
            while (1) {
                sink_ret = av_buffersink_get_frame(buffersink_ctx, filt_frame);
                if (sink_ret == AVERROR(EAGAIN) || sink_ret == AVERROR_EOF)
                    break;
                if (sink_ret < 0)
                    return -1;
                frame_out_cnt++;
                if (fp)
                    store_frame(filt_frame);
                av_frame_unref(filt_frame);
            }
            av_frame_unref(frame);
        }
    }
    return 0;
}

static int dec_video_flush(AVPacket *pkt, AVFrame *frame, AVFrame *filt_frame)
{
    int ret, sink_ret;

    ret = avcodec_send_packet(dec_ctx, pkt);
    if (ret < 0) {
        av_log(NULL, AV_LOG_ERROR, "Error while sending a packet to the decoder, ret %d\n", ret);
        return -1;
    }

    while (1) {
        ret = avcodec_receive_frame(dec_ctx, frame);
        if (ret == AVERROR_EOF) {
            break;
        } else if (ret < 0) {
            if (ret != AVERROR(EAGAIN)) {
                av_log(NULL, AV_LOG_ERROR, "Error while receiving a frame from the decoder\n");
                return -1;
            }
        }

        if (ret >= 0) {
            frame->pts = frame->best_effort_timestamp;

            /* push the decoded frame into the filtergraph */
            if (av_buffersrc_add_frame_flags(buffersrc_ctx, frame, AV_BUFFERSRC_FLAG_KEEP_REF) < 0) {
                av_log(NULL, AV_LOG_ERROR, "Error while feeding the filtergraph\n");
                break;
            }

            /* pull filtered frames from the filtergraph */
            while (1) {
                sink_ret = av_buffersink_get_frame(buffersink_ctx, filt_frame);
                if (sink_ret == AVERROR(EAGAIN) || sink_ret == AVERROR_EOF)
                    break;
                if (sink_ret < 0)
                    return -1;
                frame_out_cnt++;
                if (fp)
                    store_frame(filt_frame);
                av_frame_unref(filt_frame);
            }
            av_frame_unref(frame);
        }
    }

    return 0;
}

int main(int argc, char **argv)
{
    int ret;
    AVPacket packet = { .data = NULL, .size = 0 };
    AVFrame *frame;
    AVFrame *filt_frame;

    av_log_set_level(avloglevel);

    if (argc <= 2) {
        av_log(NULL, AV_LOG_ERROR,
               "Usage: %s <input file> <dec name> <output file>\n", argv[0]);
        return 1;
    }
    ret = init_vpe_device();
    if (ret < 0) {
        av_log(NULL, AV_LOG_ERROR, "Failed to init a VPE device\n");
        return -1;
    }

    dec_name = malloc(128);
    strcpy(dec_name, argv[2]);

#ifdef LATENCY_PROFILE
    in_time = malloc(NFRAMES * sizeof(struct timeval));
    if (!in_time) {
        av_log(NULL, AV_LOG_ERROR, "malloc error\n");
        return -1;
    }
    out_time = malloc(NFRAMES * sizeof(struct timeval));
    if (!out_time) {
        av_log(NULL, AV_LOG_ERROR, "malloc error\n");
        return -1;
    }
#endif

    if (argc <= 3) {
        output_file = NULL;
    } else {
        output_file = malloc(128);
        strcpy(output_file, argv[3]);
        fp = fopen(output_file, "wb");
        if (fp == NULL) {
            av_log(NULL, AV_LOG_ERROR, "Failed to open output file\n");
            return -1;
        }
    }

    if ((ret = open_input_file(argv[1])) < 0)
        goto end;
    av_log(NULL, AV_LOG_DEBUG, "open_input_file success\n");
    printf("open_input_file success\n");

    frame = av_frame_alloc();
    filt_frame = av_frame_alloc();
    if (!frame || !filt_frame) {
        perror("Could not allocate frame");
        goto end;
    }

    frame_out_cnt = 0;
    gettimeofday(&start_time, NULL);
    /* read all packets */
    while (1) {
        if (av_read_frame(fmt_ctx, &packet) < 0)
            break;
        ret = dec_video(&packet, frame, filt_frame);
        if (ret < 0)
            goto end;

        av_packet_unref(&packet);
    }

    packet.data = NULL;
    packet.size = 0;
    ret = dec_video_flush(&packet, frame, filt_frame);
    if (ret < 0) {
        goto end;
    }

    av_log(NULL, AV_LOG_INFO, "\n");

end:
    av_packet_unref(&packet);
    av_frame_free(&frame);
    av_frame_free(&filt_frame);
#ifdef LATENCY_PROFILE
    free(in_time);
    free(out_time);
#endif

    avcodec_free_context(&dec_ctx);
    avformat_close_input(&fmt_ctx);
    av_buffer_unref(&hw_device_ctx);

    free(dec_name);
    if (fp) {
        fclose(fp);
    }

    return ret ? 1 : 0;
}